
INCLUDES=['C:\\bld\\jcc_1555530291603\\_h_env\\Library\\include', 'C:\\bld\\jcc_1555530291603\\_h_env\\Library\\include\\win32']
CFLAGS=['/EHsc', '/D_CRT_SECURE_NO_WARNINGS']
DEBUG_CFLAGS=['/Od', '/DDEBUG']
LFLAGS=['/DLL', '/LIBPATH:C:\\bld\\jcc_1555530291603\\_h_env\\Library\\lib', 'Ws2_32.lib', 'jvm.lib']
IMPLIB_LFLAGS=['/IMPLIB:%s']
SHARED=True
VERSION="3.5"
